package dev.pfaff.altargui.internal.render;

import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.GL;
import dev.pfaff.glass.opengl.GLDebugUtil;
import dev.pfaff.glass.opengl.GLVersion;
import dev.pfaff.glass.OS;
import dev.pfaff.glass.opengl.MutPointer;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.Linker;
import java.lang.foreign.SymbolLookup;
import java.lang.invoke.MethodHandles;

public final class GLFeatures {
	public final GLVersion glVersion;
	private final @Nullable KHR_debug khrDebugImpl;

	public GLFeatures(GL gl) {
		glVersion = GLVersion.get(gl);
		var lookup = gl.symbolLookup();
		if (lookup.find("glDebugMessageCallback").isPresent()) {
			khrDebugImpl = new KHR_debug() {
				@Override
				public void glDebugMessageCallback(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
					gl.glDebugMessageCallback(callback, userParam);
				}
			};
		} else if (lookup.find("glDebugMessageCallbackKHR").isPresent()) {
			khrDebugImpl = new KHR_debug() {
				@Override
				public void glDebugMessageCallback(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
					gl.glDebugMessageCallbackKHR(callback, userParam);
				}
			};
		} else if (lookup.find("glDebugMessageCallbackARB").isPresent()) {
			khrDebugImpl = new KHR_debug() {
				@Override
				public void glDebugMessageCallback(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
					gl.glDebugMessageCallbackARB(callback, userParam);
				}
			};
		} else if (lookup.find("glDebugMessageCallbackAMD").isPresent()) {
			// there's no way they would write to the user param, right?
			khrDebugImpl = new KHR_debug() {
				@Override
				public void glDebugMessageCallback(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
					gl.glDebugMessageCallbackAMD(callback, new MutPointer<>(userParam.ms()));
				}
			};
		} else {
			khrDebugImpl = null;
		}
	}

	public boolean canUseGlProgramUniformFunctions() {
		// SIGSEGV when using these functions on macOS, so don't.
		if (OS.HOST instanceof OS.macOS) return false;
		return this.glVersion.major() > 4 || (this.glVersion.major() == 4 && this.glVersion.minor() >= 1);
	}

	public @Nullable KHR_debug getKHR_Debug() {
		return khrDebugImpl;
	}
}
