package dev.pfaff.altargui.pathfinder.fontkit;

import dev.pfaff.altargui.pathfinder.RectF;
import dev.pfaff.altargui.pathfinder.RectI;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SegmentAllocator;

import static dev.pfaff.altargui.pathfinder.Pathfinder.*;
import static java.lang.foreign.ValueLayout.JAVA_FLOAT;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public record FKFontRef(MemorySegment ptr) {
	public FKMetrics metrics() {
		try (var arena = Arena.ofConfined()) {
			var metrics = (MemorySegment) MH_FKFontMetrics.invokeExact((SegmentAllocator) arena, ptr());
			return new FKMetrics((int) VH_FKMetrics_units_per_em.get(metrics),
					(float) VH_FKMetrics_ascent.get(metrics),
					(float) VH_FKMetrics_descent.get(metrics),
					(float) VH_FKMetrics_line_gap.get(metrics),
					(float) VH_FKMetrics_underline_position.get(metrics),
					(float) VH_FKMetrics_underline_thickness.get(metrics),
					(float) VH_FKMetrics_cap_height.get(metrics),
					(float) VH_FKMetrics_x_height.get(metrics),
					RectF.fromCorners((float) VH_FKMetrics_bounding_box_origin_x.get(metrics),
							(float) VH_FKMetrics_bounding_box_origin_y.get(metrics),
							(float) VH_FKMetrics_bounding_box_lower_right_x.get(metrics),
							(float) VH_FKMetrics_bounding_box_lower_right_y.get(metrics)));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	@Nullable
	public Integer glyphForCodepoint(int codepoint) {
		try {
			var result = (long) MH_FKFontGlyphForCodepoint.invokeExact(ptr(), codepoint);
			if ((result >>> 32) != 0) {
				return (int) result;
			} else {
				return null;
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void rasterizeCodepoint(MemorySegment canvasPixelBuffer,
								   int canvasWidth, int canvasHeight,
								   FKCanvasFormat canvasFormat,
								   int glyphId,
								   float pointSize,
								   float translationX, float translationY,
								   FKCanvasRasterizationOptions rasterizationOptions) {
		try (var arena = Arena.ofConfined()) {
			var canvasSize = arena.allocateFrom(JAVA_INT, canvasWidth, canvasHeight);
			var translation = arena.allocateFrom(JAVA_FLOAT, translationX, translationY);
			var result = (int) MH_FKFontRasterizeGlyph.invokeExact(ptr(),
					canvasPixelBuffer,
					canvasPixelBuffer.byteSize(),
					canvasSize,
					canvasFormat.value(),
					glyphId,
					pointSize,
					translation,
					rasterizationOptions.value());
			if (result != 0) {
				throw new RuntimeException();
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public RectI rasterBounds(int glyphId, float pointSize, FKCanvasRasterizationOptions options) {
		try (var arena = Arena.ofConfined()) {
			var dst = arena.allocate(ML_PFRectI);
			var result = (int) MH_FKFontRasterBounds.invokeExact(ptr(), glyphId, pointSize, options.value(), dst);
			if (result != 0) throw new RuntimeException();
			var originX = dst.getAtIndex(JAVA_INT, 0);
			var originY = dst.getAtIndex(JAVA_INT, 1);
			var lowerRightX = dst.getAtIndex(JAVA_INT, 2);
			var lowerRightY = dst.getAtIndex(JAVA_INT, 3);
			return RectI.fromCorners(originX, originY, lowerRightX, lowerRightY);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
